# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from ..core import ScalarField, VectorField, TemporalScalarFields,\
    TemporalVectorFields
from ..utils.types import ARRAYTYPES, NUMBERTYPES, STRINGTYPES
from ..utils import make_unit
from ..field_treatment import get_gradients
import numpy as np
from scipy import linalg


def get_angle_deviation(vectorfield, radius=None, ind=False, mask=None,
                        raw=False, local_treatment='none', order=1):
    """
    Return the angle deviation field.

    Parameters
    ----------
    vectorfield : VectorField object
        .
    radius : number, optionnal
        The radius used to choose the zone where to compute
        for each field oint. If not mentionned, a value is choosen in
        ordre to have about 8 points in the circle. It allow to get good
        result, without big computation cost.
    ind : boolean
        If 'True', radius is expressed on number of vectors.
        If 'False' (default), radius is expressed on axis unit.
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        gamma will be compute only where mask is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    local_treatment : string, optional
        If 'None' (default), angles are taken directly from the velocity field
        If 'galilean_inv', angles are taken from localy averaged velocity field
        if 'local', angles are taken from velocity fields where the velocity of
        the central point is localy substracted.
    order : number, optional
        Order used to compute the deviation
        (default 1 for sum of differences, 2 for standart deviation (std)
        or more)
    """
    # Checking parameters coherence
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if radius is None:
        radius = 1.9
        ind = True
    if not isinstance(radius, NUMBERTYPES):
        raise TypeError("'radius' must be a number")
    if not isinstance(ind, bool):
        raise TypeError("'ind' must be a boolean")
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'zone' must be an array of boolean")
    else:
        mask = np.array(mask)
    if not isinstance(raw, bool):
        raise TypeError()
    if not isinstance(local_treatment, STRINGTYPES):
        raise TypeError()
    if local_treatment not in ['none', 'galilean_inv', 'local']:
        raise ValueError()
    if not isinstance(order, NUMBERTYPES):
        raise TypeError()
    if order < 1:
        raise ValueError()
    # Getting data
    theta = vectorfield.theta
    mask, nmbpts, mask_dev, mask_border, mask_surr, motif =\
        _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                            dev_pass=False)
    # Computing criterion
    # creating reference dispersion functions
    best_fun = np.array([2.*np.pi/nmbpts]*nmbpts)
    worse_fun = np.array([0]*(nmbpts - 1) + [2.*np.pi])
    worse_value = (np.sum(np.abs(worse_fun - best_fun)**order))**(1./order)
    # Loop on points
    deviation = np.zeros(vectorfield.shape)
    for inds, pos, _ in vectorfield:
        ind_x = inds[0]
        ind_y = inds[1]
        # stop if masked or on border or with a masked surrouinding point
        if mask[ind_x, ind_y] or mask_surr[ind_x, ind_y]\
                or mask_border[ind_x, ind_y]:
            continue
        # getting neighbour points
        indsaround = motif + inds
        # getting neighbour angles  (galilean inv or not)
        if local_treatment == 'none':
            angles = theta[indsaround[:, 0], indsaround[:, 1]]
        elif local_treatment == 'galilean_inv':
            tmp_Vx = vectorfield.comp_x[indsaround[:, 0], indsaround[:, 1]]
            tmp_Vy = vectorfield.comp_y[indsaround[:, 0], indsaround[:, 1]]
            tmp_Vx -= np.mean(tmp_Vx)
            tmp_Vy -= np.mean(tmp_Vy)
            angles = _get_angles(tmp_Vx, tmp_Vy)
        elif local_treatment == 'local':
            tmp_Vx = vectorfield.comp_x[indsaround[:, 0], indsaround[:, 1]]
            tmp_Vy = vectorfield.comp_y[indsaround[:, 0], indsaround[:, 1]]
            tmp_Vx -= vectorfield.comp_x[ind_x, ind_y]
            tmp_Vy -= vectorfield.comp_x[ind_x, ind_y]
            angles = _get_angles(tmp_Vx, tmp_Vy)
        # getting neightbour angles repartition
        angles = np.sort(angles)
        d_angles = np.empty(angles.shape)
        d_angles[0:-1] = angles[1::] - angles[:-1:]
        d_angles[-1] = angles[0] + 2*np.pi - angles[-1]
        # getting neighbour angles deviation
        deviation[ind_x, ind_y] = (1 - (np.sum(np.abs(d_angles - best_fun) **
                                               order)) **
                                   (1./order)/worse_value)
    # Applying masks
    mask = np.logical_or(mask, mask_border)
    mask = np.logical_or(mask, mask_surr)
    # Creating gamma ScalarField
    if raw:
        return np.ma.masked_array(deviation, mask)
    else:
        deviation_sf = ScalarField()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        deviation_sf.import_from_arrays(axe_x, axe_y, deviation, mask,
                                        unit_x=unit_x, unit_y=unit_y,
                                        unit_values=make_unit(''))
        return deviation_sf


def get_gamma(vectorfield, radius=None, ind=False, kind='gamma1', mask=None,
              raw=False, dev_pass=False):
    """
    Return the gamma scalar field. Gamma criterion is used in
    vortex analysis.
    The fonction recognize if the field is ortogonal, and use an
    apropriate algorithm.

    Parameters
    ----------
    vectorfield : VectorField or TemporalVectorFields object
        .
    radius : number, optionnal
        The radius used to choose the zone where to compute
        gamma for each point. If not mentionned, a value is choosen in
        ordre to have about 8 points in the circle. It allow to get good
        result, without big computation cost.
    ind : boolean
        If 'True', radius is expressed on number of vectors.
        If 'False' (default), radius is expressed on axis unit.
    kind : string
        If 'gamma1' (default), compute gamma1 criterion.
        If 'gamma1b', compute gamma1 criterion with velocity corrector.
        (multiply with the mean velocity)
        If 'gamma2', compute gamma2 criterion (with relative velocities)
        If 'gamma2b', compute gamma2 criterion with a velocity corrector.
        (hide uniform velocity zone)
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        gamma will be compute only where mask is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    dev_pass : boolean, optional
        If 'True', the algorithm compute gamma criterion only where the
        velocity angles deviation is strong (faster if there is few points).
        Work only with 'gamma1'
    """
    # Checking parameters coherence
    if not isinstance(vectorfield, (VectorField, TemporalVectorFields)):
        raise TypeError()
    if radius is None:
        radius = 1.9
        ind = True
    if not isinstance(radius, NUMBERTYPES):
        raise TypeError("'radius' must be a number")
    if not isinstance(ind, bool):
        raise TypeError("'ind' must be a boolean")
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    if not isinstance(kind, STRINGTYPES):
        raise TypeError("'kind' must be a string")
    if kind not in ['gamma1', 'gamma2', 'gamma1b', 'gamma2b']:
        raise ValueError("Unkown value for kind")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'zone' must be an array of boolean")
    else:
        mask = np.array(mask)
    if kind in ['gamma2', 'gamma2b']:
        dev_pass = False
    # if TemporalVectorFields
    if isinstance(vectorfield, TemporalVectorFields):
        gammas = TemporalScalarFields()
        for time, field in zip(vectorfield.times, vectorfield.fields):
            gamma = get_gamma(field, radius=radius, ind=ind, kind=kind,
                              mask=mask, raw=raw, dev_pass=dev_pass)
            gammas.add_field(gamma, time=time,
                             unit_times=vectorfield.unit_times)
        return gammas
    # getting data and masks
    Vx = vectorfield.comp_x
    Vy = vectorfield.comp_y
    norm_v = vectorfield.magnitude
    mask, nmbpts, mask_dev, mask_border, mask_surr, motif =\
        _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                            dev_pass)
    # getting the vectors between center and neighbouring
    deltax = axe_x[1] - axe_x[0]
    deltay = axe_y[1] - axe_y[0]
    vector_a_x = np.zeros(motif.shape[0])
    vector_a_y = np.zeros(motif.shape[0])
    for i, indaround in enumerate(motif):
        vector_a_x[i] = indaround[0]*deltax
        vector_a_y[i] = indaround[1]*deltay
    norm_vect_a = (vector_a_x**2 + vector_a_y**2)**(.5)
    # Loop on points
    gammas = np.zeros(vectorfield.shape)
    for inds, pos, _ in vectorfield:
        ind_x = inds[0]
        ind_y = inds[1]
        # stop if masked or on border or with a masked surrouinding point
        if mask[ind_x, ind_y] or mask_surr[ind_x, ind_y]\
                or mask_border[ind_x, ind_y] or mask_dev[ind_x, ind_y]:
            continue
        # getting neighbour points
        indsaround = motif + inds
        # If necessary, compute mean velocity on points (gamma2)
        v_mean = [0., 0.]
        v_mean2 = [0., 0.]
        fact = 1
        if kind in ['gamma1b', 'gamma2', 'gamma2b']:
            v_mean[0] = np.mean(Vx[indsaround[:, 0], indsaround[:, 1]])
            v_mean[1] = np.mean(Vy[indsaround[:, 0], indsaround[:, 1]])
        if kind in ['gamma2b']:
            v_mean2[0] = np.mean((Vx[indsaround[:, 0], indsaround[:, 1]] -
                                 v_mean[0])**2)
            v_mean2[1] = np.mean((Vy[indsaround[:, 0], indsaround[:, 1]] -
                                 v_mean[1])**2)
            fact = np.sqrt(v_mean2[0] + v_mean2[1]) / \
                np.sqrt(v_mean[0]**2 + v_mean[1]**2)
            if np.abs(fact) > 1:
                fact = 1.
        # Loop on neighbouring points
        gamma = 0.
        for i, indaround in enumerate(indsaround):
            inda_x = indaround[0]
            inda_y = indaround[1]
            # getting vectors for scalar product
            if kind in ['gamma1', 'gamma1b']:
                vector_b_x = Vx[inda_x, inda_y]
                vector_b_y = Vy[inda_x, inda_y]
                denom = norm_v[inda_x, inda_y]*norm_vect_a[i]
                if denom != 0:
                    gamma += (vector_a_x[i]*vector_b_y -
                              vector_a_y[i]*vector_b_x)/denom
            elif kind in ['gamma2', 'gamma2b']:
                vector_b_x = Vx[inda_x, inda_y] - v_mean[0]
                vector_b_y = Vy[inda_x, inda_y] - v_mean[1]
                denom = (vector_b_x**2 + vector_b_y**2)**.5*norm_vect_a[i]
                if denom != 0:
                    gamma += (vector_a_x[i]*vector_b_y -
                              vector_a_y[i]*vector_b_x)/denom
        # adapting with factors
        if kind in ['gamma1', 'gamma2']:
            gamma = gamma/nmbpts
        elif kind == 'gamma1b':
            gamma = gamma/nmbpts*np.sqrt(v_mean[0]**2 + v_mean[1]**2)
        elif kind == 'gamma2b':
            gamma = gamma/nmbpts*fact
        # storing computed gamma value
        gammas[ind_x, ind_y] = gamma
    # Applying masks
    mask = np.logical_or(mask, mask_border)
    mask = np.logical_or(mask, mask_surr)
    # Creating gamma ScalarField
    if raw:
        return np.ma.masked_array(gammas, mask)
    else:
        gamma_sf = ScalarField()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        gamma_sf.import_from_arrays(axe_x, axe_y, gammas, mask,
                                    unit_x=unit_x, unit_y=unit_y,
                                    unit_values=make_unit(''))
        return gamma_sf


def get_NL_residual_vorticity(vectorfield, radius=None, ind=False, mask=None,
                              raw=False):
    """
    Return the residual vorticity computed with non-local gradients.

    Parameters
    ----------
    vectorfield : VectorField object
        .
    radius : number, optionnal
        The radius used to choose the zone where to compute
        gamma for each point. If not mentionned, a value is choosen in
        ordre to have about 8 points in the circle. It allow to get good
        result, without big computation cost.
    ind : boolean
        If 'True', radius is expressed on number of vectors.
        If 'False' (default), radius is expressed on axis unit.
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        gamma will be compute only where mask is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    """
    # Checking parameters coherence
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if radius is None:
        radius = 1.9
        ind = True
    if not isinstance(radius, NUMBERTYPES):
        raise TypeError("'radius' must be a number")
    if not isinstance(ind, bool):
        raise TypeError("'ind' must be a boolean")
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'zone' must be an array of boolean")
    else:
        mask = np.array(mask)
    # getting data and masks
    Vx = vectorfield.comp_x
    Vy = vectorfield.comp_y
    mask, nmbpts, mask_dev, mask_border, mask_surr, motif =\
        _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                            dev_pass=False)
    # Loop on points to get non-local gradients
    Exx = np.zeros(vectorfield.shape, dtype=float)
    Exy = np.zeros(vectorfield.shape, dtype=float)
    Eyx = np.zeros(vectorfield.shape, dtype=float)
    Eyy = np.zeros(vectorfield.shape, dtype=float)
    for inds, pos, _ in vectorfield:
        ind_x = inds[0]
        ind_y = inds[1]
        # stop if masked or on border or with a masked surrouinding point
        if mask[ind_x, ind_y] or mask_surr[ind_x, ind_y]\
                or mask_border[ind_x, ind_y] or mask_dev[ind_x, ind_y]:
            continue
        # getting neighbour points
        indsaround = motif + inds
        # non-local gradients computation by linear fitting
        ind_xs = indsaround[:, 0]
        ind_ys = indsaround[:, 1]
        Vxs = Vx[ind_xs, ind_ys]
        Vys = Vy[ind_xs, ind_ys]
        Exx[ind_x, ind_y], a_xx = np.polyfit(axe_x[ind_xs], Vxs, 1)
        Exy[ind_x, ind_y], a_xy = np.polyfit(axe_y[ind_ys], Vxs, 1)
        Eyx[ind_x, ind_y], _ = np.polyfit(axe_x[ind_xs], Vys, 1)
        Eyy[ind_x, ind_y], _ = np.polyfit(axe_y[ind_ys], Vys, 1)
    # getting principal rate of strain (s)
    s = np.sqrt(4*Exx**2 + (Exy + Eyx)**2)/2.
    # getting the vorticity-tensor component
    omega = (Eyx - Exy)/2.
    omega_abs = np.abs(omega)
    sign_omega = np.sign(omega)
    sign_omega[sign_omega == 0] = 1
    filt = s < omega_abs
    # getting the residual vorticity
    res_vort = np.zeros(vectorfield.shape)
    res_vort[filt] = sign_omega[filt]*(omega_abs[filt] - s[filt])
    # Applying masks
    mask = np.logical_or(mask, mask_border)
    mask = np.logical_or(mask, mask_surr)
    # Creating gamma ScalarField
    if raw:
        return np.ma.masked_array(res_vort, mask)
    else:
        gamma_sf = ScalarField()
        unit_values = vectorfield.unit_values/vectorfield.unit_x
        res_vort *= unit_values.asNumber()
        unit_values /= unit_values.asNumber()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        gamma_sf.import_from_arrays(axe_x, axe_y, res_vort, mask,
                                    unit_x=unit_x, unit_y=unit_y,
                                    unit_values=unit_values)
        return gamma_sf


def get_kappa(vectorfield, radius=None, ind=False, kind='kappa1', mask=None,
              raw=False, dev_pass=False):
    """
    Return the kappa scalar field. Kappa criterion is used in
    vortex analysis.
    The fonction recognize if the field is ortogonal, and use an
    apropriate algorithm.

    Parameters
    ----------
    vectorfield : VectorField object
    radius : number, optionnal
        The radius used to choose the zone where to compute
        kappa for each point. If not mentionned, a value is choosen in
        ordre to have about 8 points in the circle. It allow to get good
        result, without big computation cost.
    ind : boolean
        If 'True', radius is expressed on number of vectors.
        If 'False' (default), radius is expressed on axis unit.
    kind : string
        If 'kappa1' (default), compute kappa1 criterion.
        If 'kappa2', compute kappa2 criterion (with relative velocities).
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        kappa will be compute only where mask is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    dev_pass : boolean, optional
        If 'True', the algorithm compute gamma criterion only where the
        velocity angles deviation is strong (faster if there is few points)
    """
    # Checking parameters coherence
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if radius is None:
        radius = 1.9
        ind = True
    if not isinstance(radius, NUMBERTYPES):
        raise TypeError("'radius' must be a number")
    if not isinstance(ind, bool):
        raise TypeError("'ind' must be a boolean")
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    if not isinstance(kind, STRINGTYPES):
        raise TypeError("'kind' must be a string")
    if kind not in ['kappa1', 'kappa2']:
        raise ValueError("Unkown value for kind")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'zone' must be an array of boolean")
    else:
        mask = np.array(mask)
    # getting data and masks
    Vx = vectorfield.comp_x
    Vy = vectorfield.comp_y
    norm_v = vectorfield.magnitude
    mask, nmbpts, mask_dev, mask_border, mask_surr, motif =\
        _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                            dev_pass)
    # getting the vectors between center and neighbouring
    deltax = axe_x[1] - axe_x[0]
    deltay = axe_y[1] - axe_y[0]
    vector_a_x = np.zeros(motif.shape[0])
    vector_a_y = np.zeros(motif.shape[0])
    for i, indaround in enumerate(motif):
        vector_a_x[i] = indaround[0]*deltax
        vector_a_y[i] = indaround[1]*deltay
    norm_vect_a = (vector_a_x**2 + vector_a_y**2)**(.5)
    # Loop on points
    kappas = np.zeros(vectorfield.shape)
    for inds, pos, _ in vectorfield:
        ind_x = inds[0]
        ind_y = inds[1]
        # stop if masked or on border or with a masked surrouinding point
        if mask[ind_x, ind_y] or mask_surr[ind_x, ind_y]\
                or mask_border[ind_x, ind_y] or mask_dev[ind_x, ind_y]:
            continue
        # getting neighbour points
        indsaround = motif + np.array(inds)
        # If necessary, compute mean velocity on points (kappa2)
        v_mean = [0., 0.]
        if kind == 'kappa2':
            nmbpts = len(indsaround)
            for indaround in indsaround:
                v_mean[0] += Vx[ind_x, ind_y]
                v_mean[1] += Vy[ind_x, ind_y]
            v_mean[0] /= nmbpts
            v_mean[1] /= nmbpts
        # Loop on neighbouring points
        kappa = 0.
        nmbpts = len(indsaround)
        for i, indaround in enumerate(indsaround):
            inda_x = indaround[0]
            inda_y = indaround[1]
            # getting vectors for scalar product
            vector_b_x = Vx[inda_x, inda_y] - v_mean[0]
            vector_b_y = Vy[inda_x, inda_y] - v_mean[1]
            if kind == 'kappa1':
                denom = norm_v[inda_x, inda_y]*norm_vect_a[i]
            else:
                denom = (vector_b_x**2 + vector_b_y**2)**.5*norm_vect_a[i]
            # getting scalar product
            if denom != 0:
                kappa += (vector_a_x[i]*vector_b_x +
                          vector_a_y[i]*vector_b_y)/denom
        # storing computed kappa value
        kappas[ind_x, ind_y] = kappa/nmbpts
    # Applying masks
    mask = np.logical_or(mask, mask_border)
    mask = np.logical_or(mask, mask_surr)
    # Creating kappa ScalarField
    if raw:
        return np.ma.masked_array(kappas, mask)
    else:
        kappa_sf = ScalarField()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        kappa_sf.import_from_arrays(axe_x, axe_y, kappas, mask,
                                    unit_x=unit_x, unit_y=unit_y,
                                    unit_values=make_unit(''))
        return kappa_sf


def get_iota(vectorfield, mask=None, radius=None, ind=False, raw=False):
    """
    Return the iota scalar field. iota criterion is used in
    vortex analysis.
    The fonction is only usable on orthogonal fields.
    Warning : This function is minimum at the saddle point center, and
    maximum around this point.

    Parameters
    ----------
    vectorfield : VectorField object
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        iota2 will be compute only where zone is 'False'.
    radius : number, optionam
        If specified, the velocity field is smoothed with gaussian filter
        of the given radius before computing the vectors angles.
    ind : boolean, optional
        If 'True', radius is an indice number, if 'False', radius if in the
        field units (default).
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    """
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask)
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    # smoothing if necessary and getting theta
    if radius is not None:
        if not ind:
            dx = axe_x[1] - axe_x[0]
            dy = axe_y[1] - axe_y[0]
            radius = radius/((dx + dy)/2.)
            ind = True
        tmp_vf = vectorfield.copy()
        tmp_vf.smooth(tos='gaussian', size=radius, inplace=True)
        theta = tmp_vf.theta
        mask = np.logical_or(mask, tmp_vf.mask)
    else:
        theta = vectorfield.theta
        mask = np.logical_or(mask, vectorfield.mask)
    # calcul du gradients de theta
    # necesary steps to avoid big gradients by passing from 0 to 2*pi
    theta1 = theta.copy()
    theta2 = theta.copy()
    theta2[theta2 > np.pi] -= 2*np.pi
    theta1_x, theta1_y = np.gradient(theta1)
    theta2_x, theta2_y = np.gradient(theta2)
    filtx = np.abs(theta1_x) > np.abs(theta2_x)
    filty = np.abs(theta1_y) > np.abs(theta2_y)
    theta_x = theta1_x.copy()
    theta_x[filtx] = theta2_x[filtx]
    theta_y = theta1_y.copy()
    theta_y[filty] = theta2_y[filty]
    iota = 1/2.*np.sqrt(theta_x**2 + theta_y**2)
    # getting mask
    maskf = np.logical_or(vectorfield.mask, np.isnan(iota))
    # returning
    if raw:
        return np.ma.masked_array(iota, maskf)
    else:
        iota_sf = ScalarField()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        iota_sf.import_from_arrays(axe_x, axe_y, iota, maskf,
                                   unit_x=unit_x, unit_y=unit_y,
                                   unit_values=make_unit(''))
        return iota_sf


def get_enstrophy(vectorfield, radius=None, ind=False, mask=None,
                  raw=False):
    """
    Return the enstriphy field.

    Parameters
    ----------
    vectorfield : VectorField object
        .
    radius : number, optionnal
        The radius used to choose the zone where to integrate
        enstrophy for each point. If not mentionned, a value is choosen in
        ordre to have about 8 points in the circle.
    ind : boolean
        If 'True', radius is expressed on number of vectors.
        If 'False' (default), radius is expressed on axis unit.
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        gamma will be compute only where mask is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    """
    # Checking parameters coherence
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if radius is None:
        radius = 1.9
        ind = True
    if not isinstance(radius, NUMBERTYPES):
        raise TypeError("'radius' must be a number")
    if not isinstance(ind, bool):
        raise TypeError("'ind' must be a boolean")
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask)
    # getting data and masks
    vort2 = get_vorticity(vectorfield, raw=False)**2
    unit_values = vort2.unit_values*vectorfield.unit_x*vectorfield.unit_y
    vort2 = vort2.values
    mask, nmbpts, mask_dev, mask_border, mask_surr, motif =\
        _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                            dev_pass=False)
    dv = ((vectorfield.axe_x[1] - vectorfield.axe_x[0]) *
          (vectorfield.axe_y[1] - vectorfield.axe_y[0]))
    # Loop on points
    enstrophy = np.zeros(vectorfield.shape)
    for inds, pos, _ in vectorfield:
        ind_x = inds[0]
        ind_y = inds[1]
        # stop if masked or on border or with a masked surrouinding point
        if mask[ind_x, ind_y] or mask_surr[ind_x, ind_y]\
                or mask_border[ind_x, ind_y]:
            continue
        # getting neighbour points
        indsaround = motif + inds
        # Loop on neighbouring points
        loc_enstrophy = 0.
        for i, indaround in enumerate(indsaround):
            loc_enstrophy += vort2[indaround[0], indaround[1]]
        # storing computed gamma value
        enstrophy[ind_x, ind_y] = loc_enstrophy*dv
    # Applying masks
    mask = np.logical_or(mask, mask_border)
    mask = np.logical_or(mask, mask_surr)
    # Creating gamma ScalarField
    if raw:
        return np.ma.masked_array(enstrophy, mask)
    else:
        enstrophy_sf = ScalarField()
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        scale = unit_values.asNumber()
        enstrophy *= scale
        unit_values = unit_values/scale
        enstrophy_sf.import_from_arrays(axe_x, axe_y, enstrophy, mask,
                                        unit_x=unit_x, unit_y=unit_y,
                                        unit_values=unit_values)
        return enstrophy_sf


def _non_local_criterion_precomputation(vectorfield, mask, radius, ind,
                                        dev_pass):
    """
    """
    # Importing data from vectorfield (velocity, axis and mask)
    mask = np.logical_or(mask, vectorfield.mask)
    # Compute motif and motif angles on an arbitrary point
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    indcentral = [int(len(axe_x)/2.), int(len(axe_y)/2.)]
    if ind:
        motif = vectorfield.get_points_around(indcentral, radius, ind)
        motif = motif - indcentral
        motif = np.delete(motif, int(len(motif)/2), axis=0)
    else:
        ptcentral = [axe_x[indcentral[0]], axe_y[indcentral[1]]]
        motif = vectorfield.get_points_around(ptcentral, radius, ind)
        motif = motif - indcentral
    nmbpts = len(motif)
    # Generating masks
    # creating surrounding masked point zone mask
    mask_surr = np.zeros(mask.shape)
    inds_masked = np.transpose(np.where(mask))
    for ind_masked in inds_masked:
        for i, j in motif + ind_masked:
            # continue if outside the field
            if i < 0 or j < 0 or i >= mask_surr.shape[0]\
                    or j >= mask_surr.shape[1]:
                continue
            mask_surr[i, j] = True
    # creating near-border zone mask
    if ind:
        indx = np.arange(len(axe_x))
        indy = np.arange(len(axe_y))
        border_x = np.logical_or(indx <= indx[0] + (int(radius) - 1),
                                 indx >= indx[-1] - (int(radius) - 1))
        border_y = np.logical_or(indy <= indy[0] + (int(radius) - 1),
                                 indy >= indy[-1] - (int(radius) - 1))
        border_x, border_y = np.meshgrid(border_x, border_y)
        mask_border = np.transpose(np.logical_or(border_x, border_y))
    else:
        delta = (axe_x[1] - axe_x[0] + axe_y[1] - axe_y[0])/2
        border_x = np.logical_or(axe_x <= axe_x[0] + (radius - delta),
                                 axe_x >= axe_x[-1] - (radius - delta))
        border_y = np.logical_or(axe_y <= axe_y[0] + (radius - delta),
                                 axe_y >= axe_y[-1] - (radius - delta))
        border_x, border_y = np.meshgrid(border_x, border_y)
        mask_border = np.transpose(np.logical_or(border_x, border_y))
    # creating dev mask
    mask_dev = np.zeros(vectorfield.shape)
    if dev_pass:
        dev = get_angle_deviation(vectorfield, radius=radius, ind=ind,
                                  raw=True)
        mask_dev = dev < 0.1
    # returning
    return (mask, nmbpts, mask_dev, mask_border, mask_surr,
            motif)


def _get_angles(Vx, Vy, check=False):
    """
    Return the angles from velocities vectors.
    """
    if check:
        # check parameters
        if not isinstance(Vx, ARRAYTYPES):
            raise TypeError()
        Vx = np.array(Vx)
        if not Vx.ndim == 1:
            raise ValueError()
        if not isinstance(Vy, ARRAYTYPES):
            raise TypeError()
        Vy = np.array(Vy)
        if not Vy.ndim == 1:
            raise ValueError()
        if not Vx.shape == Vy.shape:
            raise ValueError()
    # get data
    norm = (Vx**2 + Vy**2)**(.5)
    # getting angle
    theta = np.arccos(Vx/norm)
    theta[Vy < 0] = 2*np.pi - theta[Vy < 0]
    return theta


def get_divergence(vf, raw=False):
    """
    Return a scalar field with the 2D divergence.

    Parameters
    ----------
    vf : VectorField or TemporalVectorfields
        Field(s) on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.

    Returns
    -------
    div : ScalarField or TemporalScalarFields
        Divergence field(s)
    """
    if isinstance(vf, VectorField):
        tmp_vf = vf.fill(inplace=False)
        axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
        comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
        mask = tmp_vf.mask
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        Exx, _ = np.gradient(comp_x, dx, dy)
        _, Eyy = np.gradient(comp_y, dx, dy)
        div = Exx + Eyy
        if raw:
            return div
        else:
            unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
            unit_values = vf.unit_values/vf.unit_x
            div *= unit_values.asNumber()
            unit_values /= unit_values.asNumber()
            div_sf = ScalarField()
            div_sf.import_from_arrays(axe_x, axe_y, div, mask=mask,
                                      unit_x=unit_x, unit_y=unit_y,
                                      unit_values=unit_values)
            return div_sf
    elif isinstance(vf, TemporalVectorFields):
        if raw:
            div_tsf = np.empty((len(vf.fields), vf.shape[0], vf.shape[1]),
                               dtype=float)
            for i, field in enumerate(vf.fields):
                div_tsf[i] = get_divergence(field, raw=True)
        else:
            div_tsf = TemporalScalarFields()
            for i, field in enumerate(vf.fields):
                tmp_div = get_divergence(field, raw=False)
                div_tsf.add_field(tmp_div, time=vf.times[i],
                                  unit_times=vf.unit_times)
        return div_tsf
    else:
        raise TypeError()


def get_vorticity(vf, raw=False):
    """
    Return a scalar field with the z component of the vorticity.

    Parameters
    ----------
    vf : VectorField or TemporalVectorfields
        Field(s) on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.

    Returns
    -------
    vort : ScalarField or TemporalScalarFields
        Vorticity field(s)
    """
    if isinstance(vf, VectorField):
        tmp_vf = vf.fill(inplace=False)
        axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
        comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
        mask = tmp_vf.mask
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        _, Exy = np.gradient(comp_x, dx, dy)
        Eyx, _ = np.gradient(comp_y, dx, dy)
        vort = (Eyx - Exy)/2.
        if raw:
            return vort
        else:
            unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
            unit_values = vf.unit_values/vf.unit_x
            vort *= unit_values.asNumber()
            unit_values /= unit_values.asNumber()
            vort_sf = ScalarField()
            vort_sf.import_from_arrays(axe_x, axe_y, vort, mask=mask,
                                       unit_x=unit_x, unit_y=unit_y,
                                       unit_values=unit_values)
            return vort_sf
    elif isinstance(vf, TemporalVectorFields):
        if raw:
            vort_tsf = np.empty((len(vf.fields), vf.shape[0], vf.shape[1]),
                                dtype=float)
            for i, field in enumerate(vf.fields):
                vort_tsf[i] = get_vorticity(field, raw=True)
        else:
            vort_tsf = TemporalScalarFields()
            for i, field in enumerate(vf.fields):
                tmp_vort = get_vorticity(field, raw=False)
                vort_tsf.add_field(tmp_vort, time=vf.times[i],
                                   unit_times=vf.unit_times)
        return vort_tsf
    else:
        raise TypeError()


def get_stokes_vorticity(vf, window_size=2, raw=False):
    """
    Return a scalar field with the z component of the vorticity using
    Stokes' theorem.

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    window_size : integer, optional
        Window size for stokes approximation of the vorticity.
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.

    Notes
    -----
    Seal et al., “Quantitative characteristics of a laminar,
    unsteady necklace vortex system at a rectangular block-flat plate
    juncture,” Journal of Fluid Mechanics, vol. 286, pp. 117–135, 1995.

    """
    # getting data
    axe_x, axe_y = vf.axe_x, vf.axe_y
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    Vx = vf.comp_x
    Vy = vf.comp_y
    mask = vf.mask
    # creating new axis
    new_axe_x = np.arange(np.mean(axe_x[0:window_size]),
                          np.mean(axe_x[-window_size::] + dx*.9),
                          dx)
    new_axe_y = np.arange(np.mean(axe_y[0:window_size]),
                          np.mean(axe_y[-window_size::] + dy*.9),
                          dy)
    # Loop on field
    vort = np.zeros((len(new_axe_x), len(new_axe_y)))
    new_mask = np.zeros((len(new_axe_x), len(new_axe_y)), dtype=bool)
    for i in np.arange(len(axe_x) - window_size + 1):
        for j in np.arange(len(axe_y) - window_size + 1):
            # reinitialazing
            tmp_vort = 0.
            # checking masked values
            if np.any(mask[i:i + window_size, j:j + window_size]):
                new_mask[i, j] = True
                continue
            # summing over first border
            bord_vec = -Vy[i, j:j + window_size].copy()
            tmp_vort += np.trapz(bord_vec, dx=dy)
            # summing over second border
            bord_vec = Vy[i + window_size - 1, j:j + window_size].copy()
            tmp_vort += np.trapz(bord_vec, dx=dy)
            # summing over third border
            bord_vec = Vx[i:i + window_size, j].copy()
            tmp_vort += np.trapz(bord_vec, dx=dx)
            # summing over fourth border
            bord_vec = -Vx[i:i + window_size, j + window_size - 1].copy()
            tmp_vort += np.trapz(bord_vec, dx=dx)
            # adding coefficients
            tmp_vort *= 1./(dx*dy*window_size**2)
            # storing
            vort[i, j] = tmp_vort
    # returning
    if raw:
        return vort
    else:
        unit_values = vf.unit_values/vf.unit_x
        vort *= unit_values.asNumber()
        unit_values /= unit_values.asNumber()
        vort_sf = ScalarField()
        vort_sf.import_from_arrays(new_axe_x, new_axe_y, vort, mask=new_mask,
                                   unit_x=vf.unit_x, unit_y=vf.unit_y,
                                   unit_values=unit_values)
        return vort_sf


def get_swirling_strength(vf, raw=False):
    """
    Return a scalar field with the swirling strength
    (imaginary part of the eigenvalue of the velocity Jacobian)

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.

    Notes
    -----
    Zhou, J., R. J. Adrian, S. Balachandar, et T. M. Kendall.
    « Mechanisms for generating coherent packets of hairpin vortices in
    channel flow ». Journal of Fluid Mechanics 387 (mai 1999): 353‑96.

    """
    if not isinstance(vf, VectorField):
        raise TypeError()
    tmp_vf = vf.copy()
    tmp_vf.fill()
    # Getting gradients and axes
    axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
    mask = tmp_vf.mask
    du_dx, du_dy, dv_dx, dv_dy = get_gradients(vf, raw=True)
    # swirling stregnth matrix
    swst = np.zeros(tmp_vf.shape)
    # loop on  points
    for i in np.arange(len(axe_x)):
        for j in np.arange(len(axe_y)):
            if not mask[i, j]:
                lapl = [[du_dx[i, j], du_dy[i, j]],
                        [dv_dx[i, j], dv_dy[i, j]]]
                eigvals = np.linalg.eigvals(lapl)
                swst[i, j] = np.max(np.imag(eigvals))
    mask = np.logical_or(mask, np.isnan(swst))
    # creating ScalarField object
    if raw:
        return swst
    else:
        unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
        # TODO: Implement unities
        unit_values = ""
        tmp_sf = ScalarField()
        tmp_sf.import_from_arrays(axe_x, axe_y, swst, mask=mask,
                                  unit_x=unit_x, unit_y=unit_y,
                                  unit_values=unit_values)
        return tmp_sf


def get_improved_swirling_strength(vf, raw=False):
    """
    Return a scalar field with the improved swirling strength

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.

    Notes
    -----
    Chakraborty, Pinaki, S. Balachandar, et Ronald J. Adrian.
    « On the Relationships between Local Vortex Identification Schemes ».
    Journal of Fluid Mechanics 535 (5 juillet 2005): 189‑214.

    """
    if not isinstance(vf, VectorField):
        raise TypeError()
    tmp_vf = vf.copy()
    tmp_vf.fill()
    # Getting gradients and axes
    axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
    comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
    mask = tmp_vf.mask
    dx = axe_x[1] - axe_x[0]
    dy = axe_y[1] - axe_y[0]
    du_dx, du_dy = np.gradient(comp_x, dx, dy)
    dv_dx, dv_dy = np.gradient(comp_y, dx, dy)
    # swirling stregnth matrix
    swst = np.zeros(tmp_vf.shape)
    # loop on  points
    for i in np.arange(len(axe_x)):
        for j in np.arange(len(axe_y)):
            if not mask[i, j]:
                lapl = [[du_dx[i, j], du_dy[i, j]],
                        [dv_dx[i, j], dv_dy[i, j]]]
                eigvals = np.linalg.eigvals(lapl)
                lambcr = np.real(eigvals[0])
                lambci = np.abs(np.imag(eigvals[0]))
                if lambci == 0:
                    mask[i, j] = True
                swst[i, j] = lambcr/lambci
    mask = np.logical_or(mask, np.isnan(swst))
    # creating ScalarField object
    if raw:
        return swst
    else:
        unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
        # TODO: Implement unities
        unit_values = ""
        tmp_sf = ScalarField()
        tmp_sf.import_from_arrays(axe_x, axe_y, swst, mask=mask,
                                  unit_x=unit_x, unit_y=unit_y,
                                  unit_values=unit_values)
        return tmp_sf


def get_q_criterion(vectorfield, mask=None, raw=False):
    """
    Return the scalar field of the 2D Q criterion .
    Define as "1/2*(R**2 - S**2)" , with "R" the deformation tensor,
    norm and "S" the rate of rotation tensor norm.

    Parameters
    ----------
    vectorfield : VectorField object
    mask : array of boolean, optional
        Has to be an array of the same size of the vector field object,
        Q criterion will be compute only where zone is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    """
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask)
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    # calcul des gradients
    Exx, Exy, Eyx, Eyy = get_gradients(vectorfield, raw=True)
    # calcul de Qcrit
    norm_rot2 = 1/2.*(Exy - Eyx)**2
    norm_shear2 = (Exx**2 + 1./2.*(Exy + Eyx)**2 + Eyy**2)
    qcrit = .5*(norm_rot2) - norm_shear2
    unit_values = (vectorfield.unit_values/vectorfield.unit_x)**2
    scale = unit_values.asNumber()
    qcrit *= scale
    unit_values = unit_values/scale
    if raw:
        return np.ma.masked_array(qcrit, mask)
    else:
        q_sf = ScalarField()
        q_sf.import_from_arrays(axe_x, axe_y, qcrit, mask,
                                unit_x=vectorfield.unit_x,
                                unit_y=vectorfield.unit_y,
                                unit_values=unit_values)
        return q_sf


def get_Nk_criterion(vectorfield, mask=None, raw=False):
    """
    Return the scalar field of the 2D Nk criterion .
    Define as "||Omega||/||S||" , with "||Omega||" the rotation rate tensor
    norm and ||S|| the shear rate tensor norm.

    Parameters
    ----------
    vectorfield : VectorField object
    mask : array of boolean, optional
        Has to be an array of the same size of the vector field object,
        Nk criterion will be compute only where zone is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.

    Notes
    -----
    See J. Jeong and F. Hussain, “On the identification of a vortex,” Journal
    of Fluid Mechanics, vol. 285, pp. 69–94, 1995.

    """
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask)
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    # calcul des gradients
    Exx, Exy, Eyx, Eyy = get_gradients(vectorfield, raw=True)
    # calcul de Nk
    norm_rot = 1./2.**.5*np.abs(Exy - Eyx)
    norm_shear = (Exx**2 + 1./2.*(Exy + Eyx)**2 + Eyy**2)**.5
    Nkcrit = norm_rot/norm_shear
    unit_values = make_unit('')
    if raw:
        return np.ma.masked_array(Nkcrit, mask)
    else:
        q_sf = ScalarField()
        q_sf.import_from_arrays(axe_x, axe_y, Nkcrit, mask,
                                unit_x=vectorfield.unit_x,
                                unit_y=vectorfield.unit_y,
                                unit_values=unit_values)
        return q_sf


def get_delta_criterion(vectorfield, mask=None, raw=False):
    """
    Return the scalar field of the 2D Delta criterion .
    Define as "(Q/3)**3 + (R/2)**2" , with "Q" the Q criterion,
    and "R" the determinant of the jacobian matrice of the velocity.

    Parameters
    ----------
    vectorfield : VectorField object
    mask : array of boolean, optional
        Has to be an array of the same size of the vector field object,
        iota2 will be compute only where zone is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.

    Note
    ----
    Negative values of Delta mean that the local streamline pattern is closed
    or spiraled.
    """
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if mask is None:
        mask = np.zeros(vectorfield.shape, dtype=bool)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask, dtype=bool)
    axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
    # calcul des gradients
    Exx, Exy, Eyx, Eyy = get_gradients(vectorfield, raw=True)
    # calcul de Q
    norm_rot2 = 1/2.*(Exy - Eyx)**2
    norm_shear2 = (Exx**2 + 1./2.*(Exy + Eyx)**2 + Eyy**2)
    Q = .5*(norm_rot2) - norm_shear2
    # calcul de R
    R = np.zeros(Exx.shape)
    for i in np.arange(Exx.shape[0]):
        for j in np.arange(Exx.shape[1]):
            Jac = [[Exx[i, j], Exy[i, j]], [Eyx[i, j], Eyy[i, j]]]
            eigval, eigvect = np.linalg.eig(Jac)
            if np.all(np.imag(eigval) == 0):
                mask[i, j] = True
                continue
            R[i, j] = -np.linalg.det(Jac)
    # calcul de Delta
    delta = (Q/3.)**3 + (R/2.)**2
    unit_values = ((vectorfield.unit_values/vectorfield.unit_x)**2)**3
    scale = unit_values.asNumber()
    delta *= scale
    unit_values = unit_values/scale
    if raw:
        return np.ma.masked_array(delta, mask)
    else:
        delta_sf = ScalarField()
        delta_sf.import_from_arrays(axe_x, axe_y, delta, mask,
                                    unit_x=vectorfield.unit_x,
                                    unit_y=vectorfield.unit_y,
                                    unit_values=unit_values)
        return delta_sf


def get_lambda2(vectorfield, mask=None, raw=False):
    """
    Return the lambda2 scalar field. According to ... vortex are defined by
    zone of negative values of lambda2.
    The fonction is only usable on orthogonal fields.

    Parameters
    ----------
    vectorfield : VectorField object
    mask : array of boolean, optionnal
        Has to be an array of the same size of the vector field object,
        iota2 will be compute only where zone is 'False'.
    raw : boolean, optional
        If 'False' (default), a ScalarField is returned,
        if 'True', an array is returned.
    """
    # check parameter
    if not isinstance(vectorfield, VectorField):
        raise TypeError("'vectorfield' must be a VectorField object")
    if mask is None:
        mask = np.zeros(vectorfield.shape)
    elif not isinstance(mask, ARRAYTYPES):
        raise TypeError("'mask' must be an array of boolean")
    else:
        mask = np.array(mask)
    mask = np.logical_or(mask, vectorfield.mask)
    # getting velocity gradients
    Udx, Udy, Vdx, Vdy = get_gradients(vectorfield, raw=True)
    mask = np.logical_or(mask, Udx.mask)
    # creating returning matrix
    lambda2 = np.zeros(vectorfield.shape)
    # loop on points
    for i in np.arange(lambda2.shape[0]):
        for j in np.arange(lambda2.shape[1]):
            # check if masked
            if mask[i, j]:
                continue
            # getting symmetric and antisymetric parts
            S = 1./2.*np.array([[2*Udx[i, j], Udy[i, j] + Vdx[i, j]],
                                [Vdx[i, j] + Udy[i, j], 2*Vdy[i, j]]])
            Omega = 1./2.*np.array([[0, Udy[i, j] - Vdx[i, j]],
                                    [Vdx[i, j] - Udy[i, j], 0]])
            # getting S^2 + Omega^2
            M = np.dot(S, S) + np.dot(Omega, Omega)
            # getting second eigenvalue
            lambds = linalg.eig(M, left=False, right=False)
            l2 = np.min(np.real(lambds))
            # storing lambda2
            lambda2[i, j] = l2
    # returning
    if raw:
        return np.ma.masked_array(l2, mask)
    else:
        lambd_sf = ScalarField()
        axe_x, axe_y = vectorfield.axe_x, vectorfield.axe_y
        unit_x, unit_y = vectorfield.unit_x, vectorfield.unit_y
        lambd_sf.import_from_arrays(axe_x, axe_y, lambda2, mask,
                                    unit_x=unit_x, unit_y=unit_y,
                                    unit_values=make_unit(''))
        return lambd_sf


def get_residual_vorticity(vf, raw=False):
    """
    Return a scalar field with the residual of the vorticity.
    (see Kolar (2007)).

    Parameters
    ----------
    vf : VectorField or Velocityfield
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.
    """
    if isinstance(vf, VectorField):
        # getting data
        tmp_vf = vf.copy()
        tmp_vf.fill(inplace=True, reduce_tri=True)
        axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
        comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
        mask = tmp_vf.mask
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        # getting gradients
        Exx, Exy = np.gradient(comp_x, dx, dy)
        Eyx, Eyy = np.gradient(comp_y, dx, dy)
        # getting principal rate of strain (s)
        s = np.sqrt(4*Exx**2 + (Exy + Eyx)**2)/2.
        # getting the vorticity-tensor component
        omega = (Eyx - Exy)/2.
        omega_abs = np.abs(omega)
        sign_omega = np.zeros(omega.shape, dtype=int)
        sign_omega[omega_abs == 0] = 1.
        sign_omega[omega_abs != 0] = (omega[omega_abs != 0] /
                                      omega_abs[omega_abs != 0])
        filt = s < omega_abs
        # getting the residual vorticity
        res_vort = np.zeros(tmp_vf.shape)
        res_vort[filt] = sign_omega[filt]*(omega_abs[filt] - s[filt])
        if raw:
            return res_vort
        else:
            unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
            unit_values = vf.unit_values/vf.unit_x
            res_vort *= unit_values.asNumber()
            unit_values /= unit_values.asNumber()
            vort_sf = ScalarField()
            vort_sf.import_from_arrays(axe_x, axe_y, res_vort, mask=mask,
                                       unit_x=unit_x, unit_y=unit_y,
                                       unit_values=unit_values)
            return vort_sf
    elif isinstance(vf, TemporalVectorFields):
        if raw:
            tvf = []
            for field in vf.fields:
                tvf.append(get_residual_vorticity(field, raw=True))
        else:
            tvf = TemporalScalarFields()
            for i, field in enumerate(vf.fields):
                tvf.add_field(get_residual_vorticity(field, raw=False),
                              time=vf.times[i], unit_times=vf.unit_times)
        # return
        return tvf
    else:
        raise TypeError()


def get_shear_vorticity(vf, raw=False):
    """
    Return a scalar field with the shear vorticity.
    (see Kolar (2007)).

    Parameters
    ----------
    vf : VectorField or VectorFields
        Field on which compute shear stress
    raw : boolean, optional
        If 'True', return an arrays,
        if 'False' (default), return a ScalarField object.
    """
    if isinstance(vf, VectorField):
        # getting data
        tmp_vf = vf.copy()
        tmp_vf.crop_masked_border()
        tmp_vf.fill(inplace=True, reduce_tri=True)
        axe_x, axe_y = tmp_vf.axe_x, tmp_vf.axe_y
        comp_x, comp_y = tmp_vf.comp_x, tmp_vf.comp_y
        mask = tmp_vf.mask
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        # getting gradients
        Exx, Exy = np.gradient(comp_x, dx, dy)
        Eyx, Eyy = np.gradient(comp_y, dx, dy)
        # getting principal rate of strain (s)
        s = np.sqrt(4*Exx**2 + (Exy + Eyx)**2)/2.
        # getting the vorticity-tensor component
        omega = (Eyx - Exy)/2.
        omega_abs = np.abs(omega)
        sign_omega = omega/omega_abs
        filt1 = s < omega_abs
        filt2 = np.logical_not(filt1)
        # getting the residual vorticity
        sh_vort = np.zeros(vf.shape)
        sh_vort[filt2] = omega[filt2]
        sh_vort[filt1] = sign_omega[filt1]*(s[filt1])
        if raw:
            return sh_vort
        else:
            unit_x, unit_y = tmp_vf.unit_x, tmp_vf.unit_y
            unit_values = vf.unit_values/vf.unit_x
            sh_vort *= unit_values.asNumber()
            unit_values /= unit_values.asNumber()
            vort_sf = ScalarField()
            vort_sf.import_from_arrays(axe_x, axe_y, sh_vort, mask=mask,
                                       unit_x=unit_x, unit_y=unit_y,
                                       unit_values=unit_values)
            return vort_sf
    elif isinstance(vf, TemporalVectorFields):
        ret = TemporalScalarFields()
        for i, field in enumerate(vf.fields):
            tmp_ret = get_shear_vorticity(field)
            ret.add_field(tmp_ret, time=vf.times[i],
                          unit_times=vf.unit_times)
        return ret
    else:
        raise TypeError()



import IMTreatment.file_operation as imtio
import numpy as np


def sane_parameters():
    pass


def parametric_test(func, kwargs, update=False):
    alphabet = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l",
                "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x",
                "y", "z"]
    fun_name = func.__name__
    for let, kwarg in zip(alphabet, kwargs):
        filename = "test_{}_{}.cimt".format(fun_name, let)
        res = func(**kwarg)
        # Update if necessary
        if update:
            imtio.export_to_file(res, filename)
        # If the file is not present, create it
        try:
            res2 = imtio.import_from_file(filename)
        except FileNotFoundError:
            print("file '{}' is not present (normal for a first run),"
                  "I created it for you !".format(filename))
            imtio.export_to_file(res, filename)
            continue
        # Else try recursively to test for equality
        try:
            res[0][0]
            for r, r2 in zip(res, res2):
                assert np.all(r == r2)
        except TypeError:
            try:
                res[0]
                assert np.all(res == res2)
            except TypeError:
                assert res == res2



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import scipy.ndimage.measurements as msr
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import simps

from ..core import Points, Profile, ScalarField
from ..utils import ProgressCounter, make_unit
from ..vortex_criterions import (get_NL_residual_vorticity,
                                 get_residual_vorticity, get_vorticity)


def get_vortex_radius(VF, vort_center, NL_radius=None, eps_detection=0.1,
                      output_center=False, output_unit=False):
    """
    Return the radius of the given vortex, use the residual vorticity.

    Parameters
    ----------
    VF : vectorfield object
        Velocity field on which compute gamma2.
    vort_center : 2x1 array
        Approximate position of the vortex center.
    NL_radius : number, optional
        IF specified, radius used for the non-local computation of the
        gradients.
    eps_detection : number
        epsilon used to determine the edge of the vortex (default is 0.1 for
        10% of the maximum vorticity value).
    output_center : boolean, optional
        If 'True', return the associated vortex center, computed using center
        of mass algorythm.
    output_unit ; boolean, optional
        If 'True', return the associated unit.

    Returns
    -------
    radius : number
        Average radius of the vortex. If no vortex is found, 0 is returned.
    center : 2x1 array of numbers
        If 'output_center' is 'True', contain the newly computed vortex center.
    unit_radius : Unit object
        Radius unity
    """

    # getting data
    if NL_radius is None:
        vort = get_residual_vorticity(VF)
    else:
        vort = get_NL_residual_vorticity(VF, radius=NL_radius, ind=False,
                                         mask=None, raw=False)
    dx = VF.axe_x[1] - VF.axe_x[0]
    dy = VF.axe_y[1] - VF.axe_y[0]
    # getting the ~initial maximum vorticity of the zone
    ind_x = VF.get_indice_on_axe(1, vort_center[0], kind='nearest')
    ind_y = VF.get_indice_on_axe(2, vort_center[1], kind='nearest')
    max_vort = vort.values[ind_x, ind_y]
    # getting zone around point
    zones, nmb_zones = msr.label(vort.values > eps_detection*max_vort)
    zone_number = zones[ind_x, ind_y]
    # getting the real maximum vorticity
    max_vort = np.max(vort.values[zones == zone_number])
    # getting the real zones
    zones, nmb_zones = msr.label(vort.values > eps_detection*max_vort)
    zone_number = zones[ind_x, ind_y]
    # getting the zone area and radius
    area = np.sum(zones == zone_number)*dx*dy
    radius = (area/np.pi)**.5
    # getting the new center
    if output_center:
        center = np.array(msr.center_of_mass(np.abs(vort.values),
                                             zones == zone_number))
        center[0] = VF.axe_x[0] + center[0]*dx
        center[1] = VF.axe_y[0] + center[1]*dy
    # optional computed unit
    if output_unit:
        unit_radius = (VF.unit_x**2 + VF.unit_y**2)**.5
    # return
    if not output_unit and not output_center:
        return radius
    elif output_unit and not output_center:
        return radius, unit_radius
    elif not output_unit and output_center:
        return radius, center
    else:
        return radius, center, unit_radius


def get_vortex_radius_time_evolution(TVFS, traj, NL_radius=None,
                                     eps_detection=0.1,
                                     output_center=False, verbose=False):
    """
    Return the radius evolution in time for the given vortex center trajectory.

    Use the criterion $|gamma2| > 2/pi$. The returned radius is an average value
    if the vortex zone is not circular.

    Parameters
    ----------
    TVFS : TemporalField object
        Velocity field on which compute gamma2.
    traj : Points object
        Trajectory of the vortex.
    NL_radius : number, optional
        IF specified, radius used for the non-local computation of the
        gradients.
    eps_detection : number
        epsilon used to determine the edge of the vortex (default is 0.1 for
        10% of the maximum vorticity value).
    output_center : boolean, optional
        If 'True', return a Points object with associated vortex centers,
        computed using center of mass algorythm.
    verbose : boolean
        .

    Returns
    -------
    radius : Profile object
        Average radius of the vortex. If no vortex is found, 0 is returned.
    center : Points object
        If 'output_center' is 'True', contain the newly computed vortex center.
    """
    radii = np.empty((len(traj.xy),))
    if verbose:
        pg = ProgressCounter(init_mess="Begin vortex radii detection",
                             nmb_max=len(traj.xy),
                             name_things='fields',
                             perc_interv=1)
    # computing with vortex center
    if output_center:
        centers = Points(unit_x=TVFS.unit_x, unit_y=TVFS.unit_y,
                         unit_v=TVFS.unit_times)

        for i, pt in enumerate(traj):
            if verbose:
                pg.print_progress()
            # getting time and associated velocity field
            time = traj.v[i]
            field = TVFS.fields[TVFS.times == time][0]
            # getting radius and center
            rad, cent = get_vortex_radius(field, traj.xy[i],
                                          NL_radius=NL_radius,
                                          eps_detection=eps_detection,
                                          output_center=True)
            radii[i] = rad
            centers.add(cent, time)
    # computing without vortex centers
    else:
        for i, _ in enumerate(traj):
            if verbose:
                pg.print_progress()
            # getting time and associated velocity field
            time = traj.v[i]
            field = TVFS.fields[TVFS.times == time][0]
            # getting radius
            radii[i] = get_vortex_radius(field, traj.xy[i],
                                         NL_radius=NL_radius,
                                         eps_detection=eps_detection,
                                         output_center=False)
    # returning
    mask = radii == 0.
    radii_prof = Profile(traj.v, radii, mask=mask, unit_x=TVFS.unit_times,
                         unit_y=TVFS.unit_x)
    if output_center:
        return radii_prof, centers
    else:
        return radii_prof


def get_vortex_property(VF, vort_center, size_crit=None, size_crit_lim=0.1,
                        prop_crit=None, output_unit=False, verbose=False):
    """
    Return a property of a particular vortex.

    Parameters
    ----------
    VF : vectorfield object
        Base velocity field.
    vort_center : 2x1 array
        Approximate position of the vortex center.
    size_crit : function or 'value'
        Function applied to 'VF' and returning a ScalarField used to get the
        vortex area.
        (Default is residual vorticity)
        If 'value', only the value at the given point is returned.
    size_crit_lim : number
        Used to determine the size criterion interval defining the vortex area
        (i.e. the vortex area is the area around the vortex center where
        the size criterion is superior to 'size_crit_lim' times the value at
        the center)
        (Default is 0.1 (10%))
        Useless if 'size_crit='value'
    prop_crit : function
        Function applied to 'VF' and returning a ScalarField used to get the
        property value (Default is residual vorticity)
    output_unit : boolean, optional
        If 'True', return the associated unit.
    verbose : bool
        If 'True', display information and graph along computation.

    Returns
    -------
    prop : number
        Property associated to the vortex.
        (Is the integral of 'prop_crit' result on the area defined by
        'size_crit')
    """
    # default behavior
    if size_crit is None:
        size_crit = get_residual_vorticity
    if prop_crit is None:
        prop_crit = get_residual_vorticity
    # getting data
    if size_crit == "value":
        prop_crit = prop_crit(VF)
        val = prop_crit.get_value(*vort_center)
        if output_unit:
            return val, prop_crit.unit_values
        else:
            return val
    if size_crit == prop_crit:
        prop_crit = size_crit(VF)
        size_crit = np.abs(prop_crit)
    else:
        size_crit = np.abs(size_crit(VF))
        prop_crit = prop_crit(VF)
    ind_x = VF.get_indice_on_axe(1, vort_center[0], kind='nearest')
    ind_y = VF.get_indice_on_axe(2, vort_center[1], kind='nearest')
    unit_int = prop_crit.unit_values*prop_crit.unit_x*prop_crit.unit_y
    # check if value is positive at the vortex center
    if VF.magnitude[ind_x, ind_y] <= 0:
        raise ValueError()
    # get first guess vortex zone (use value at center as a predica of
    #    the maxima)
    tmp_maxi = size_crit.values[ind_x, ind_y]
    vort_zones = size_crit.values > tmp_maxi*size_crit_lim
    vort_zones_labs, _ = msr.label(vort_zones)
    lab = vort_zones_labs[ind_x, ind_y]
    # if we're outside, just give up
    if lab == 0:
        if output_unit:
            return 0, unit_int
        else:
            return 0
    # get final vortex zone with real maxima
    size_crit_maxi = np.max(size_crit.values[lab == vort_zones_labs])
    vort_zones = size_crit.values > size_crit_maxi*size_crit_lim
    vort_zones_labs, _ = msr.label(vort_zones)
    lab = vort_zones_labs[ind_x, ind_y]
    # get prop_crit integral along zone
    tmp_prop_crit = prop_crit.values.copy()
    tmp_prop_crit[lab != vort_zones_labs] = 0.
    dx = VF.axe_x[1] - VF.axe_x[0]
    dy = VF.axe_y[1] - VF.axe_y[0]
    prop = simps(simps(tmp_prop_crit, dx=dy), dx=dx)
    # verbose
    if verbose:
        tmp_vort_zone = ScalarField()
        tmp_vort_zone.import_from_arrays(VF.axe_x, VF.axe_y,
                                         values=lab == vort_zones_labs)
        fig, axs = plt.subplots(2, 1)
        plt.sca(axs[0])
        size_crit.display()
        VF.display(kind='stream', color='w')
        tmp_vort_zone._display(kind='contour', levels=[-1e30, 0.5, 1e30],
                               colors='r')
        plt.plot([], color='r', label='Detected vortex zone')
        plt.plot(*vort_center, marker='o', linestyle='none', mec='k', mfc='w',
                 label="Vortex center")
        plt.title("Vortex size detection")
        plt.legend()
        plt.sca(axs[1])
        prop_crit.display()
        VF.display(kind='stream', color='w')
        tmp_vort_zone._display(kind='contour', levels=[-1e30, 0.5, 1e30],
                               colors='r')
        plt.plot([], color='r', label='Detected vortex zone')
        plt.plot(*vort_center, marker='o', linestyle='none', mec='k', mfc='w',
                 label="Vortex center")
        plt.title("Vortex property integration on vortex zone\nProp={:.2f} {}"
                  .format(prop, unit_int.strUnit()))
        plt.legend()
    # returning
    if output_unit:
        return prop, unit_int
    else:
        return prop


def get_vortex_property_time_evolution(TVFs, vort_center_traj, size_crit=None,
                                       size_crit_lim=0.1,
                                       prop_crit=None, output_unit=False,
                                       verbose=0):
    """
    Return a property of a particular vortex.

    Parameters
    ----------
    TVFs : TemporalVectorFields object
        Base velocity fields.
    vort_center : Points object
        Approximate position of the vortex centers along times.
    size_crit : function or 'value'
        Function applied to 'VF' and returning a ScalarField used to get the
        vortex area.
        (Default is residual vorticity)
        if 'value', only return the value at point.
    size_crit_lim : number
        Used to determine the size criterion interval defining the vortex area
        (i.e. the vortex area is the area around the vortex center where
        the size criterion is superior to 'size_crit_lim' times the value at
        the center)
        (Default is 0.1 (10%))
        Useless if "size_crit='value".
    prop_crit : function
        Function applied to 'VF' and returning a ScalarField used to get the
        property value (Default is residual vorticity)
    verbose : integer
        specified the number of fields to verbosify.
        Default is 0.

    Returns
    -------
    prop : Profile object
        Evolution of the property associated with the vortex long time.
    """
    # prepare storage
    times = []
    props = []
    if verbose == 1:
        field_to_verbosify = [len(vort_center_traj)/2]
    elif verbose == 2:
        field_to_verbosify = [len(vort_center_traj)/3,
                              len(vort_center_traj)*2/3]
    else:
        field_to_verbosify = np.round(np.linspace(0, len(vort_center_traj),
                                                  verbose))
    # loop on fields
    for i in range(len(TVFs)):
        # pass if vortex center is not defined for this time
        if TVFs.times[i] not in vort_center_traj.v:
            continue
        ind_traj = np.where(vort_center_traj.v == TVFs.times[i])[0][0]
        # verbosify (or not...)
        if len(times) in field_to_verbosify:
            verbose = True
        else:
            verbose = False
        # get the wanted property
        field = TVFs.fields[i]
        vc = vort_center_traj.xy[ind_traj]
        prop, unit = get_vortex_property(VF=field, vort_center=vc,
                                         size_crit=size_crit,
                                         size_crit_lim=size_crit_lim,
                                         prop_crit=prop_crit, verbose=verbose,
                                         output_unit=True)
        times.append(TVFs.times[i])
        props.append(prop)
    # store on a Profile object
    prof_prop = Profile(x=times, y=props, unit_x=TVFs.unit_times,
                        unit_y=unit)
    # returning
    return prof_prop


def get_vortex_circulation(VF, vort_center, epsilon=0.1, output_unit=False,
                           verbose=False):
    """
    Return the circulation of the given vortex.

    $\Gamma = \int_S \omega dS$
    avec : $S$ : surface su vortex ($| \omega | > \epsilon$)

    Recirculation is representative of the swirling strength.

    Warning : integral on complex domain is complex (you don't say?),
    here is just implemented a sum of accessible values on the domain.

    Parameters
    ----------
    VF : vectorfield object
        Velocity field on which compute gamma2.
    vort_center : 2x1 array
        Approximate position of the vortex center.
    epsilon : float, optional
        Relative seuil for the vorticity integral (default is 0.1).
    output_unit : boolean, optional
        If 'True', circulation unit is returned.

    Returns
    -------
    circ : float
        Vortex virculation.
    """
    # getting data
    ind_x = VF.get_indice_on_axe(1, vort_center[0], kind='nearest')
    ind_y = VF.get_indice_on_axe(2, vort_center[1], kind='nearest')
    dx = VF.axe_x[1] - VF.axe_x[0]
    dy = VF.axe_y[1] - VF.axe_y[0]
    vort = get_vorticity(VF)
    # find omega > 0.1 zones and label them
    max_vort = np.max(np.abs(vort.values[~vort.mask]))
    vort_zone = np.abs(vort.values)/max_vort > epsilon
    vort_zone, nmb_zone = msr.label(vort_zone)
    # get wanted zone label
    lab = vort_zone[ind_x, ind_y]
    # if we are outside a zone
    if lab == 0:
        if output_unit:
            return 0., make_unit("")
        else:
            return 0.
    if verbose:
        plt.figure()
        vort.display()
        plt.plot(vort_center[0], vort_center[1], 'ok')
        plt.figure()
        plt.imshow(vort_zone == lab)
    # else, we compute the circulation
    circ = np.sum(vort.values[vort_zone == lab])*dx*dy
    # if necessary, we compute the unit
    unit_circ = vort.unit_values*VF.unit_x*VF.unit_y
    circ *= unit_circ.asNumber()
    unit_circ /= unit_circ.asNumber()
    # returning
    if output_unit:
        return circ, unit_circ
    else:
        return circ



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from IMTreatment import VectorField, ScalarField, Profile
import numpy as np
from scipy import integrate
import matplotlib.pyplot as plt
import matplotlib.path as mplpath

ARRAYTYPES = (np.ndarray, list, tuple)
NUMBERTYPES = (int, int, float, complex)
STRINGTYPES = (str, str)


# object class
class System(object):
    """
    Representing a potential system (boundaries, sources, freestream, ...)
    """
    def __init__(self, u_inf=0., alpha=0.):
        """
        Parameters
        ----------
        u_inf : number, optional
            Free Stream velocity (m/s).

        """
        self.u_inf = u_inf
        self.alpha = alpha
        self.objects_0D = []
        self.objects_1D = []
        self.objects_2D = []
        self.sigma_to_refresh = False

    def add_object(self, dim, coords, kind='wall', res=1):
        """
        Add an object to the system

        Parameters
        ----------
        dim : integer
            Can be '0' for a point, '1' for a boundary or '2' for a solid.
        coords : array of number
            Coordinates of the object (1x2 array if 'dim=1', nx2 arrays else).
        kind : string, optional
            Kind of object (default is 'wall').
        res : integer
            Resolution, length of the wanted segments.
        """
        # check
        if not isinstance(dim, NUMBERTYPES):
            raise TypeError()
        if not isinstance(coords, ARRAYTYPES):
            raise TypeError()
        coords = np.array(coords)
        # object_0D
        if dim == 0:
            self.objects_0D.append(object_0D(coords, kind, res=res))
        # object_1D
        elif dim == 1:
            self.objects_1D.append(object_1D(coords, kind, res=res))
        # object_2D
        elif dim == 2:
            self.objects_2D.append(object_2D(coords, kind, res=res))
        else:
            raise ValueError()
        #
        self.sigma_to_refresh = True

    def get_solid_panels(self):
        """
        Return all the solid panels.
        """
        panels = np.array([], dtype=object)
        for obj in self.objects_1D:
            if obj.kind == 'wall':
                panels = np.append(panels, obj.panels)
        for obj in self.objects_2D:
            if obj.kind == 'wall':
                panels = np.append(panels, obj.panels)
        panels = np.array(panels, subok=True)
        return panels.flatten()

    def get_solid_paths(self):
        """
        Return the solid paths.
        """
        paths = []
        for obj in self.objects_2D:
            if obj.kind == 'wall':
                paths.append(obj.path)
        return paths

    def solving_sigma(self):
        """
        Solve sigma values for the current boundaries.
        """
        # getting solid panels
        solid_panels = self.get_solid_panels()
        nmb_solid_panels = len(solid_panels)
        # computes the source influence matrix
        A = np.empty((nmb_solid_panels, nmb_solid_panels), dtype=float)
        np.fill_diagonal(A, 0.5)
        for i, p_i in enumerate(solid_panels):
            for j, p_j in enumerate(solid_panels):
                if i == j:
                    continue
                A[i, j] = 0.5/np.pi*integral(p_i.xc, p_i.yc, p_j,
                                             p_i.cosbeta, p_i.sinbeta)
        # computes the RHS of the linear system
        b = - self.u_inf * np.cos([self.alpha - p.beta for p in solid_panels])
        # solves the linear system
        sigma = np.linalg.solve(A, b)
        for i, panel in enumerate(solid_panels):
            panel.sigma = sigma[i]
        self.sigma_to_refresh = False

        # computes the matrix of the linear system
        A = np.empty((len(solid_panels), len(solid_panels)), dtype=float)
        np.fill_diagonal(A, 0.0)
        for i, p_i in enumerate(solid_panels):
            for j, p_j in enumerate(solid_panels):
                if i != j:
                    A[i, j] = 0.5/np.pi*integral(p_i.xc, p_i.yc, p_j,
                                                 -np.sin(p_i.beta),
                                                 np.cos(p_i.beta))
        # computes the RHS of the linear system
        b = self.u_inf * np.sin([self.alpha - panel.beta
                                 for panel in solid_panels])
        # computes the tangential velocity at each panel center-point
        vt = np.dot(A, sigma) + b
        for i, panel in enumerate(solid_panels):
            panel.vt = vt[i]
#
#        ### cp ###
#        for panel in self.solid_panels:
#            panel.cp = 1.0 - (panel.vt/u_inf)**2

    def compute_velocity(self, x, y):
        """
        Return the velocity at the given point.
        """
        # update sigma if necessary
        if self.sigma_to_refresh:
            self.solving_sigma()
        solid_panels = self.get_solid_panels()
        # check if on panel
        for pan in solid_panels:
            if pan.on_panel(x, y):
                return pan.vector
        # compute u
        u = self.u_inf*np.cos(self.alpha)\
            + 0.5/np.pi*sum([p.sigma*integral(x, y, p, 1, 0)
                             for p in solid_panels])
        # compute v
        v = self.u_inf*np.sin(self.alpha)\
            + 0.5/np.pi*sum([p.sigma*integral(x, y, p, 0, 1)
                             for p in solid_panels])
        # check if velocity isinstance not too high
        if np.abs(u) > 1e30 or np.abs(v) > 1e30:
            u = 0.
            v = 0.
        return u, v

    def compute_velocity_on_grid(self, grid_x, grid_y, raw=False,
                                 remove_solid=False):
        """
        Returns the velocity field on the given grid.
        """
        # update sigma if necessary
        if self.sigma_to_refresh:
            self.solving_sigma()
        # creating the grid
        X, Y = np.meshgrid(grid_x, grid_y)
        Nx, Ny = X.shape
        u, v = np.empty((Nx, Ny), dtype=float), np.empty((Nx, Ny), dtype=float)
        # mask if we don't want velocities in the solids
        paths = self.get_solid_paths()
        mask = np.zeros((Nx, Ny), dtype=bool)
        #compute velocities
        for i in range(Nx):
            for j in range(Ny):
                x = X[i, j]
                y = Y[i, j]
                if remove_solid:
                    in_solid = np.any([path.contains_point((x, y),
                                                           radius=-1e-10)
                                       for path in paths])
                    if in_solid:
                        mask[i, j] = True
                        continue
                u_i, v_i = self.compute_velocity(x, y)
                u[i, j] = u_i
                v[i, j] = v_i
        # returning
        if raw:
            return u, v
        else:
            vf = VectorField()
            u = np.transpose(u)
            v = np.transpose(v)
            mask = np.transpose(mask)
            vf.import_from_arrays(grid_x, grid_y, u, v, mask=mask,
                                  unit_x='m', unit_y='m', unit_values='m/s')
            return vf

    def compute_velocity_on_line(self, xya, xyb, res=100, raw=False,
                                 remove_solid=False):
        """
        Return the velocity on the given line.
        """
        # update sigma if necessary
        if self.sigma_to_refresh:
            self.solving_sigma()
        # creating the line
        xa, ya = xya
        xb, yb = xyb
        x = np.linspace(xa, xb, res)
        y = np.linspace(ya, yb, res)
        vx = np.empty(x.shape)
        vy = np.empty(x.shape)
        # mask if we don't want velocities on solid
        mask = np.zeros(x.shape)
        paths = self.get_solid_paths()
        # loop on line
        for i, _ in enumerate(x):
            if remove_solid:
                in_solid = np.any([path.contains_point((x[i], y[i]))
                                   for path in paths])
                if in_solid:
                    mask[i] = True
                    continue
            v = self.compute_velocity(x[i], y[i])
            vx[i] = v[0]
            vy[i] = v[1]
        # returning
        if raw:
            return vx, vy
        else:
            x = x - np.min(x)
            y = y - np.min(y)
            new_x = np.sqrt(x**2 + y**2) - np.sqrt(x[0]**2 + y[0]**2)
            prof_vx = Profile(new_x, vx, mask=mask, unit_x='m',
                              unit_y='m/s')
            prof_vy = Profile(new_x, vy, mask=mask, unit_x='m',
                              unit_y='m/s')
            return prof_vx, prof_vy

    def compute_pressure_from_velocity(self, obj, raw=False):
        """
        Return the pressure coefficient computed from velocity data.
        """
        if isinstance(obj, VectorField):
            vx = obj.comp_x
            vy = obj.comp_y
            grid_x = obj.axe_x
            grid_y = obj.axe_y
            mask = obj.mask
            cp = 1. - (vx**2 + vy**2)/self.u_inf**2
            mask = np.logical_or(np.isnan(cp), mask)
            if raw:
                return cp
            else:
                sf = ScalarField()
                sf.import_from_arrays(grid_x, grid_y, cp, mask=mask,
                                      unit_x='m', unit_y='m', unit_values='')
                return sf
        elif isinstance(obj, (tuple, list)):
            if isinstance(obj[0], Profile):
                vx = obj[0].y
                vy = obj[1].y
                mask = np.logical_or(obj[0].mask, obj[1].mask)
                cp = 1. - (vx**2 + vy**2)/self.u_inf**2
                if raw:
                    return cp
                else:
                    prof = Profile(obj[0].x, cp, mask=mask,
                                   unit_x=obj[0].unit_x,
                                   unit_y='')
                    return prof
            else:
                raise TypeError()
        else:
            raise TypeError()

    def display(self, solid=True, panels=True):
        """
        Display the current geometry.
        """
        plt.grid(True)
        for obj in self.objects_0D:
            obj.display(panels=panels)
        for obj in self.objects_1D:
            obj.display(panels=panels)
        for obj in self.objects_2D:
            obj.display(panels=panels, solid=solid)


class object_0D(object):
    """
    Representing a Source or a sink.
    """
    def __init__(self, x, y):
        """
        Representing a Source or a sink.

        Parameters
        ----------
        x, y : number
            Coordinate of the point.
        """
        # check
        if not isinstance(x, NUMBERTYPES):
            raise TypeError()
        if not isinstance(y, NUMBERTYPES):
            raise TypeError()
        # store
        self.x = x
        self.y = y


class object_1D(object):
    """
    Representing a boundary (wall, source or sink).
    """
    def __init__(self, coords, kind, res):
        """
        Representing a solid (wall, source or sink).

        Parameters
        ----------
        coords : nx2 array of numbers
            Coordinate of the boundary path.
        kind : string
            Kind of object (default is 'wall').
        res : integer
            Resolution, number of segments on the shortest line (default=1)
        """
        # check
        if not isinstance(coords, ARRAYTYPES):
            raise TypeError()
        coords = np.array(coords, dtype=float)
        if not coords.ndim == 2:
            raise ValueError()
        if not isinstance(kind, STRINGTYPES):
            raise TypeError()
        if not isinstance(res, int):
            raise TypeError()
        if res < 1:
            raise ValueError()
        # store
        self.x = coords[:, 0]
        self.y = coords[:, 1]
        self.nmb_panels = 0
        self.panels = np.empty((1,), dtype=object)
        self.path = None
        self.kind = kind
        # refine if necessary
        self.refine(res)
        # add the panels
        self._actualize_panels()
        # actualize the path
        self._actualize_path()
        # TODO : check normals and coherence

    def _actualize_panels(self):
        # Add the coresponding panels
        self.nmb_panels = len(self.x) - 1
        self.panels = np.empty(shape=(self.nmb_panels,),
                               dtype=object)
        # creating the panels objects
        for i in np.arange(self.nmb_panels):
                self.panels[i] = Panel([self.x[i], self.y[i]],
                                       [self.x[i + 1], self.y[i + 1]])

    def _actualize_path(self):
        for i in np.arange(self.nmb_panels):
            coords = list(zip(self.x, self.y))
            self.path = mplpath.Path(coords)

    def inverse_normals(self):
        """
        Inverse normals.
        """
        self.x = self.x[::-1]
        self.y = self.y[::-1]
        self._actualize_panels()
        self._actualize_path()

    def refine(self, res):
        """
        Refine each segments of the solid.
        """
        x = self.x
        y = self.y
        new_x = []
        new_y = []
        # getting delta (length of the panels)
        lens_x = np.array(x[1::] - x[0:-1])
        lens_y = np.array(y[1::] - y[0:-1])
        lens = (lens_x**2 + lens_y**2)**.5
        min_len = np.min(lens)
        delta = min_len/float(res)
        # refining
        for i in np.arange(len(self.x) - 1):
            tmp_nmb = int(np.round(lens[i]/delta))
            if tmp_nmb <= 1:
                new_x += [x[i]]
                new_y += [y[i]]
                continue
            new_x += list(np.linspace(x[i], x[i + 1], tmp_nmb + 1)[:-1])
            new_y += list(np.linspace(y[i], y[i + 1], tmp_nmb + 1)[:-1])
        # closing
        new_x += [x[0]]
        new_y += [y[0]]
        new_x = np.array(new_x, dtype=float)
        new_y = np.array(new_y, dtype=float)
        # storing new values
        self.x = new_x
        self.y = new_y
        # updating panels ans path
        self._actualize_panels()
        self._actualize_path()

    def display(self, panels=True):
        """
        Display the 1D object
        """
        plt.plot(self.x, self.y, 'k-')
        if panels:
            for pan in self.panels:
                pan.display(wall_vel=False, nodes=True)


class object_2D(object_1D):
    """
    Representing a solid (wall, source or sink).
    """

    def __init__(self, coords, kind, res):
        """
        Representing a solid (wall, source or sink).

        Parameters
        ----------
        coords : nx2 array of numbers
            Coordinate of the boundary path.
        kind : string
            Kind of object (default is 'wall').
        res : integer
            Resolution, lenght of the wanted segments.
        """
        # check
        if not isinstance(coords, ARRAYTYPES):
            raise TypeError()
        coords = np.array(coords, dtype=float)
        if not coords.ndim == 2:
            raise ValueError()
        if not isinstance(kind, STRINGTYPES):
            raise TypeError()
        # store
        self.x = coords[:, 0]
        self.y = coords[:, 1]
        self.nmb_panels = 0
        self.panels = np.empty((1,), dtype=object)
        self.path = None
        self.kind = kind
        # close the path if necessary
        if self.x[0] != self.x[-1] or self.y[0] != self.y[-1]:
            self.x = np.append(self.x, self.x[0])
            self.y = np.append(self.y, self.y[0])
        # refine if necessary
        self.refine(res)
        # add the panels
        self._actualize_panels()
        # actualize the path
        self._actualize_path()
        # TODO : check normals and coherence

    def display(self, solid=True, panels=True):
        """
        Display the 3D object
        """
        object_1D.display(self, panels=panels)
        if solid:
            plt.fill(self.x, self.y, 'k')


# Panel class (discretization element)
class Panel(object):
    """
    Contains information related to a panel.
    """
    def __init__(self, xya, xyb, sigma=0.):
        """Initializes the panel.

        Arguments
        ---------
        xa, ya -- coordinates of the first end-point of the panel.
        xb, yb -- coordinates of the second end-point of the panel.
        """
        # checl
        self.xa = xya[0]
        self.ya = xya[1]
        self.xb = xyb[0]
        self.yb = xyb[1]
        if self.xa == self.xb and self.ya == self.yb:
            raise ValueError()
        self.xc, self.yc = (self.xa + self.xb)/2., (self.ya + self.yb)/2.
        self.length = np.sqrt((self.xb - self.xa)**2 + (self.yb - self.ya)**2)
        self.length2 = self.length**2
        # orientation of the panel (angle between x-axis and panel's normal)
        if self.xb - self.xa <= 0.:
            self.beta = np.arccos((self.yb - self.ya)/self.length)
        elif self.xb - self.xa > 0.:
            self.beta = np.pi + np.arccos(-(self.yb - self.ya)/self.length)
        self.sinbeta = np.sin(self.beta)
        self.cosbeta = np.cos(self.beta)
        self.sigma = sigma                          # source strength
        self.vt = 0.                                # tangential velocity
        self.cp = 0.                                # pressure coefficient
        # precomputation to 'on_panel' test
        if self.xa == self.xb:
            self.vertical = True
        else:
            self.vertical = False
        if self.ya == self.yb:
            self.horizontal = True
        else:
            self.horizontal = False
        if not self.vertical and not self.horizontal:
            self.coef_a = (self.yb - self.ya)/(self.xb - self.xa)
            self.coef_b = self.yb - self.coef_a*self.xb
            self.coef_C1 = (1 + self.coef_a**2)**.5

    @property
    def vector(self):
        return (-self.vt*np.sin(self.beta), self.vt*np.cos(self.beta))

    def display(self, wall_vel=True, nodes=True):
        """
        Display the panel
        """
        # plotting the panels
        plt.plot([self.xa, self.xb], [self.ya, self.yb], 'k-')
        if nodes:
            plt.plot([self.xa, self.xb], [self.ya, self.yb], 'ro')
            plt.plot([self.xc], [self.yc], 'ok')
        if wall_vel:
            vx, vy = self.vector
            plt.quiver(self.xc, self.yc, vx, vy)
        return plt.gcf()

    def on_panel(self, x, y, eps_abs=1e-4):
        """
        Check if the point is on the panel.
        """
        # check if around the point c
        dist_c = (self.xc - x)**2 + (self.yc - y)**2
        if dist_c > self.length2/2.:
            return False
        # check if on the panle line
        if self.vertical:
            dist_line = np.abs(x - self.xc)
        elif self.horizontal:
            dist_line = np.abs(y - self.yc)
        else:
            dist_line = np.abs(self.coef_a*x - y + self.coef_b)/self.coef_C1
        if dist_line < eps_abs:
            return True
        else:
            return False


#def Point(object):
#    """
#    Contains information related to a point.
#    """
#
#    def __init__(self, x, y, sigma=0.):
#        self.x = x
#        self.y = y
#        self.sigma = sigma
#
#    def display(self, **kwargs):
#        plt.plot(self.x, self.y, 'o', **kwargs)

def integral(x, y, panel, dxdz, dydz):
    """Evaluates the contribution of a panel at one point.

    Arguments
    ---------
    x, y -- Cartesian coordinates of the point.
    panel -- panel which contribution is evaluated.
    dxdz -- derivative of x in the z-direction.
    dydz -- derivative of y in the z-direction.

    Returns
    -------
    Integral over the panel of the influence at one point.
    """
    sin_beta = panel.sinbeta
    cos_beta = panel.cosbeta
    # coef de calcul (for optimization purpose)
    pre_xs = x - panel.xa
    pre_ys = y - panel.ya
    a = pre_xs**2 + pre_ys**2
    b = 2*pre_xs*sin_beta - 2*pre_ys*cos_beta
    c = sin_beta**2 + cos_beta**2

    def func(s):
        xs = pre_xs + sin_beta*s
        ys = pre_ys - cos_beta*s
        denom = a + b*s + c*s**2
        return (xs*dxdz + ys*dydz)/denom
    res = integrate.quad(lambda s: func(s), 0., panel.length, limit=50)
    if res[1] > 1e-5:
        print("integration pb, check your geometry"
              "(very small panel somewhere ?) !")
        print((x, y))
        return res[0]
    else:
        return res[0]


def get_separation_position(D=1., Vd=1., x_obst=10., theta_max=np.inf,
                            nu=1e-6, res_pot=20, res_int=500):
    """
    Use potential flow to get velocity distribution and Thwaites BL
    equation to get the separation position.

    Parameters
    ----------
    D : number
        Obstacle diameter [m].
    Vd : number
        Bulk velocity [m/s].
    x_obst : number
        Distance from CL birth to obstacle [m].
    theta_max : number
        Maximum value of BL momentum thickness (in case of confinement) [m].
    nu : number
        Kinematic viscosity [].
    res_pot : integer
        Resolution for potential flow model (default=20).
    res_int : integer
        Resolution for Thwaites integral resolution (default=500).

    Returns
    -------
    x_sep : number
        Separation position
    """
    # creating system
    syst = System(u_inf=Vd, alpha=0.)
    syst.add_object(2, [[-D/2., -D/2.], [-D/2., D/2.],
                        [D/2., D/2.], [D/2., -D/2.]],
                    kind='wall', res=res_pot)
    syst.objects_2D[0].inverse_normals()
    syst.solving_sigma()

    # getting profiles along symmetry plane
    dx = x_obst
    x_f = -D/2.
    x_0 = x_f - dx
    prof_vit = syst.compute_velocity_on_line([x_f, 0], [x_0, 0], res=res_int,
                                             remove_solid=True)[0]
    # compute theta**2
    U = prof_vit.y
    theta_0 = 0.
    x_0 = -dx
    x_f = 0.
    x = np.linspace(x_0, x_f, res_int)
    U5 = prof_vit.y**5
    integral_U = [np.trapz(U5[0:i], prof_vit.x[0:i])
                  for i in np.arange(len(prof_vit.x))]
    theta2 = theta_0**2*(U[0]/U)**6 + 0.45*nu/U**6*integral_U
    theta2[theta2 > theta_max**2] = theta_max**2
    # compute m
    deriv_U = np.gradient(U, x[1]-x[0])
    m = -theta2/nu*deriv_U
    # find separation
    m = Profile(x, m, unit_x='m', unit_y='')
    x_sep = m.get_value_position(0.09)[0]
    return -x_sep


def get_gradP_length(D=1., Vd=1., perc=0.1, nu=1e-6, res_pot=20, res_int=500,
                     eps=1e-6):
    """
    Return the position before the obstacle where the pressure is equal
    at the wanted percentage of the pressure at the obstacle,
    using potential flow theory.

    Parameters
    ----------
    D : number
        Obstacle diameter [m].
    Vd : number
        Bulk velocity [m/s].
    perc : number
        wanted percentage (default to 0.1).
    nu : number
        Kinematic viscosity [].
    res_pot : integer
        Resolution for potential flow model (default=20).
    res_int : integer
        Resolution for Thwaites integral resolution (default=500).
    eps : number
        Wanted precision on the position (for the iterative solver)
    Returns
    -------
    dP_len : number
        Position of 10% pressure.
    """
    # creating system
    syst = System(u_inf=Vd, alpha=0.)
    syst.add_object(2, [[-D/2., -D/2.], [-D/2., D/2.],
                        [D/2., D/2.], [D/2., -D/2.]],
                    kind='wall', res=res_pot)
    syst.objects_2D[0].inverse_normals()
    syst.solving_sigma()

    # iterative finder
    dx = 5*D
    x_f = -D/2.
    x_0 = x_f - dx
    res_int = 3
    P_max = 1.
    i = 0
    while True:
        i += 1
        # getting profiles along symmetry plane
        prof_vit = syst.compute_velocity_on_line([x_f, 0], [x_0, 0],
                                                 res=res_int,
                                                 remove_solid=True)
        prof_P = syst.compute_pressure_from_velocity(prof_vit)
        # get 10% pressure
        ind = np.where(prof_P.y < perc*P_max)[0]
        # check if on the good interval
        if len(ind) == 0:
            x_0 -= dx
            continue
        else:
            ind = ind[-1]
        # return if precision is achieved
        if np.abs(perc*P_max - prof_P.y[ind]) < eps:
            x_pos = -(prof_P.x[ind] + x_f)
            return x_pos
        # raise error if too man iterations
        if i > 100:
            raise Exception()
        # another run !
        x_0 = prof_P.x[ind] + x_f
        x_f = prof_P.x[ind + 1] + x_f
